# -*- coding: utf-8 -*-

import collections
import logging
from typing import Optional, Union, Tuple, Sequence, List, NamedTuple, overload, TYPE_CHECKING

from pydantic import BaseModel, Extra, validator, root_validator
import numpy as np

from ._helpers import set_module, singledispatch

if TYPE_CHECKING:
    from ._parameters import Parameters  # noqa


PointType = Sequence[int]
FloatPointType = Sequence[float]
PointSequenceType = Sequence[PointType]

InitialPointType = Tuple[int, ...]
InitialWayPointsType = Tuple[InitialPointType, ...]


MPE_MODULE = 'skmpe'
MIN_NDIM = 2


logger = logging.getLogger(MPE_MODULE)
logger.addHandler(logging.NullHandler())


class ImmutableDataObject(BaseModel):
    """Base immutable data object with validating fields
    """

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True
        allow_mutation = False


@set_module(MPE_MODULE)
class InitialInfo(ImmutableDataObject):
    """Initial info for extracting path
    """

    speed_data: np.ndarray

    start_point: InitialPointType
    end_point: InitialPointType
    way_points: InitialWayPointsType = ()

    @validator('start_point', 'end_point', 'way_points', pre=True)
    def _to_canonical(cls, v):
        return np.asarray(v).tolist()

    @root_validator
    def _check_ndim(cls, values):
        speed_data = values.get('speed_data')

        start_point = values.get('start_point')
        end_point = values.get('end_point')
        way_points = values.get('way_points')

        if (speed_data is None or
                start_point is None or
                end_point is None or
                way_points is None):
            return values

        ndim = speed_data.ndim

        if ndim < MIN_NDIM:
            raise ValueError(f'minimum dimension must be {MIN_NDIM}.')

        if ndim != len(start_point) or ndim != len(end_point):
            raise ValueError(f"the start and end points must have dimension {ndim}.")

        for point in way_points:
            if len(point) != ndim:
                raise ValueError(f"the start, end and all way points must have dimension {ndim}.")

        return values

    @root_validator
    def _check_point_duplicates(cls, values):
        start_point = values.get('start_point')
        end_point = values.get('end_point')
        way_points = values.get('way_points')

        if start_point is None or end_point is None or way_points is None:
            return values

        all_points = [start_point, end_point, *way_points]
        duplicates = [point for point, count in collections.Counter(all_points).items() if count > 1]

        if duplicates:
            raise ValueError(
                f'the points must not be duplicated, there are duplicated points: {duplicates}')

        return values

    @validator('start_point', 'end_point', 'way_points')
    def _check_points(cls, v, field, values):
        if v is None:
            return v  # pragma: no cover

        if 'speed_data' not in values:
            return v  # pragma: no cover

        speed_data = values['speed_data']

        def validate(point, index=None):
            if len(point) != speed_data.ndim:
                # this is the concern of another validator
                return point

            n_point = f' [{index}] ' if index else ' '

            for i, (n_coord, n_size) in enumerate(zip(point, speed_data.shape)):
                if n_coord < 0 or n_coord >= n_size:
                    raise ValueError(f"'{field.name}' {point}{n_point}coordinate {i} "
                                     f"is out of 'speed_data' bounds [0, {n_size}).")

            if speed_data[point] is np.ma.masked:
                raise ValueError(f"'{field.name}' {point}{n_point}inside 'speed_data' masked area.")

        if field.name == 'way_points':
            for idx, p in enumerate(v):
                validate(p, idx)
        else:
            validate(v)

        return v

    def all_points(self) -> List[PointType]:
        return [self.start_point, *self.way_points, self.end_point]

    def point_intervals(self) -> List[Tuple[PointType, PointType]]:
        all_points = self.all_points()
        return list(zip(all_points[:-1], all_points[1:]))


@set_module(MPE_MODULE)
class PathInfo(NamedTuple):
    """Extracted path info
    """

    path: np.ndarray
    start_point: PointType
    end_point: PointType
    travel_time: np.ndarray
    path_travel_times: np.ndarray
    reversed: bool


@set_module(MPE_MODULE)
class ResultPathInfo(ImmutableDataObject):
    """Result path info
    """

    path: np.ndarray
    pieces: Tuple[PathInfo, ...]

    @property
    def point_count(self) -> int:
        return self.path.shape[0]


@overload
def mpe(speed_data: np.ndarray,
        start_point: Union[PointType, np.ndarray],
        end_point: Union[PointType, np.ndarray],
        way_points: Union[PointSequenceType, np.ndarray] = (),
        *,
        parameters: Optional['Parameters'] = None) -> ResultPathInfo:
    pass  # pragma: no cover


@overload
def mpe(init_info: InitialInfo,
        *,
        parameters: Optional['Parameters'] = None) -> ResultPathInfo:
    pass  # pragma: no cover


@set_module(MPE_MODULE)
@singledispatch
def mpe(*args, **kwargs) -> ResultPathInfo:  # noqa
    """Extracts a minimal path

    Usage
    -----

    .. code-block:: python

        mpe(init_info: InitialInfo, *,
            parameters: Optional[Parameters] = None) -> ResultPathInfo

        mpe(speed_data: np.ndarray, *,
            start_point: Sequence[int],
            end_point: Sequence[int],
            way_points: Sequence[Sequence[int]] = (),
            parameters: Optional[Parameters] = None) -> ResultPathInfo

    Returns
    -------
    path_info : ResultPathInfo
        Extracted path info

    """



# -*- coding: utf-8 -*-

import contextlib
import enum

from pydantic import confloat, conint, validator

from ._base import MPE_MODULE, ImmutableDataObject
from ._helpers import set_module


@set_module(MPE_MODULE)
class TravelTimeOrder(enum.IntEnum):
    first = 1
    second = 2


@set_module(MPE_MODULE)
class OdeSolverMethod(str, enum.Enum):
    RK23 = 'RK23'
    RK45 = 'RK45'
    DOP853 = 'DOP853'
    Radau = 'Radau'
    BDF = 'BDF'
    LSODA = 'LSODA'


@set_module(MPE_MODULE)
class Parameters(ImmutableDataObject):
    """MPE algorithm parameters
    """

    travel_time_spacing: confloat(gt=0.0) = 1.0
    travel_time_order: TravelTimeOrder = TravelTimeOrder.first
    travel_time_cache: bool = False

    ode_solver_method: OdeSolverMethod = OdeSolverMethod.RK45

    integrate_time_bound: confloat(gt=0.0) = 10000.0
    integrate_min_step: confloat(ge=0.0) = 0.0
    integrate_max_step: confloat(gt=0.0) = 4.0

    dist_tol: confloat(ge=0.0) = 1e-3
    max_small_dist_steps: conint(strict=True, gt=1) = 100

    @validator('travel_time_order')
    def _check_travel_time_order(cls, v):
        if v == TravelTimeOrder.second:
            raise ValueError(
                'Currently the second order for computing travel time does not work properly.'
                '\nSee the following issue for details: https://github.com/scikit-fmm/scikit-fmm/issues/28'
            )
        return v


_default_parameters = Parameters()


@contextlib.contextmanager
def parameters(**kwargs):
    """Context manager for using specified parameters

    Parameters
    ----------

    kwargs : mapping
        The parameters

            - **travel_time_spacing** --
            - **travel_time_order** --
            - **travel_time_cache** --
            - **ode_solver_method** --
            - **integrate_time_bound** --
            - **integrate_min_step** --
            - **dist_tol** --
            - **max_small_dist_steps** --

    """

    global _default_parameters
    prev_default_parameters = _default_parameters

    _default_parameters = Parameters(**kwargs)

    try:
        yield _default_parameters
    finally:
        _default_parameters = prev_default_parameters


def default_parameters() -> Parameters:
    """Returns the default parameters

    Returns
    -------
    parameters : Parameters
        Default parameters

    """

    return _default_parameters



# -*- coding: utf-8 -*-

from typing import List

import numpy as np

from ._base import PointType, FloatPointType, MPE_MODULE
from ._helpers import set_module


@set_module(MPE_MODULE)
class MPEError(Exception):
    pass


@set_module(MPE_MODULE)
class ComputeTravelTimeError(MPEError):
    pass


@set_module(MPE_MODULE)
class PathExtractionError(MPEError):
    def __init__(self, *args,
                 travel_time: np.ndarray,
                 start_point: PointType,
                 end_point: PointType) -> None:
        super().__init__(*args)

        self.travel_time = travel_time
        self.start_point = start_point
        self.end_point = end_point


@set_module(MPE_MODULE)
class EndPointNotReachedError(PathExtractionError):
    def __init__(self, *args,
                 travel_time: np.ndarray,
                 start_point: PointType,
                 end_point: PointType,
                 extracted_points: List[FloatPointType],
                 last_distance: float,
                 reason: str) -> None:
        super().__init__(*args, travel_time=travel_time, start_point=start_point, end_point=end_point)

        self.extracted_points = extracted_points
        self.last_distance = last_distance
        self.reason = reason



# -*- coding: utf-8 -*-

import functools
import inspect


def set_module(name: str):
    """Replace __module__ for decorated object

    Parameters
    ----------
    name : str
        Module name

    Returns
    -------
    obj : Any
        The object with replaced __module__
    """

    def decorator(obj):
        obj.__module__ = name
        return obj
    return decorator


def singledispatch(func_stub):
    """Single dispatch wrapper with default implementation that raises the exception for invalid signatures
    """

    @functools.wraps(func_stub)
    @functools.singledispatch
    def _dispatch(*args, **kwargs):
        sign_args = ', '.join(f'{type(arg).__name__}' for arg in args)
        sign_kwargs = ', '.join(f'{kwname}: {type(kwvalue).__name__}' for kwname, kwvalue in kwargs.items())
        allowed_signs = ''
        i = 0

        for arg_type, func in _dispatch.registry.items():
            if arg_type is object:
                continue
            i += 1
            sign = inspect.signature(func)
            allowed_signs += f'  [{i}] => {sign}\n'

        raise TypeError(
            f"call '{func_stub.__name__}' with invalid signature:\n"
            f"  => ({sign_args}, {sign_kwargs})\n\n"
            f"allowed signatures:\n{allowed_signs}")

    return _dispatch



# -*- coding: utf-8 -*-

from typing import Optional, Union

import numpy as np

from ._base import mpe as api_dispatch
from ._base import PointType, PointSequenceType, InitialInfo, ResultPathInfo
from ._parameters import Parameters
from ._mpe import extract_path


@api_dispatch.register(np.ndarray)  # noqa
def mpe(speed_data: np.ndarray,
        start_point: Union[PointType, np.ndarray],
        end_point: Union[PointType, np.ndarray],
        way_points: Union[PointSequenceType, np.ndarray] = (),
        *,
        parameters: Optional[Parameters] = None) -> ResultPathInfo:

    init_info = InitialInfo(
        speed_data=speed_data,
        start_point=start_point,
        end_point=end_point,
        way_points=way_points
    )

    return mpe(init_info, parameters=parameters)


@api_dispatch.register(InitialInfo)  # noqa
def mpe(init_info: InitialInfo,
        *,
        parameters: Optional[Parameters] = None) -> ResultPathInfo:
    return extract_path(init_info, parameters)



# -*- coding: utf-8 -*-

from ._base import (
    InitialInfo,
    PathInfo,
    ResultPathInfo,
    mpe
)

from ._parameters import (
    TravelTimeOrder,
    Parameters,
    parameters,
    default_parameters,
)

from ._exceptions import (
    MPEError,
    ComputeTravelTimeError,
    PathExtractionError,
    EndPointNotReachedError,
)

from ._mpe import MinimalPathExtractor

# register dispatchered API
import skmpe._api as _api  # noqa


__version__ = '0.1.1'

__all__ = [
    'InitialInfo',
    'PathInfo',
    'ResultPathInfo',

    'TravelTimeOrder',
    'Parameters',
    'parameters',
    'default_parameters',

    'MPEError',
    'ComputeTravelTimeError',
    'PathExtractionError',
    'EndPointNotReachedError',

    'MinimalPathExtractor',
    'mpe',
]



# -*- coding: utf-8 -*-

import itertools
import warnings
from typing import List, Optional

import numpy as np
import skfmm as fmm

from scipy.interpolate import RegularGridInterpolator
from scipy.integrate import RK23, RK45, DOP853, Radau, BDF, LSODA
from scipy.spatial.distance import euclidean

from ._base import PointType, InitialInfo, PathInfo, ResultPathInfo, logger, MPE_MODULE
from ._helpers import set_module
from ._parameters import Parameters, OdeSolverMethod, default_parameters
from ._exceptions import ComputeTravelTimeError, PathExtractionError, EndPointNotReachedError


def make_interpolator(coords, values, fill_value: float = 0.0):
    return RegularGridInterpolator(
        coords, values, method='linear', bounds_error=False, fill_value=fill_value)


ODE_SOLVER_METHODS = {
    OdeSolverMethod.RK23: RK23,
    OdeSolverMethod.RK45: RK45,
    OdeSolverMethod.DOP853: DOP853,
    OdeSolverMethod.Radau: Radau,
    OdeSolverMethod.BDF: BDF,
    OdeSolverMethod.LSODA: LSODA,
}


@set_module(MPE_MODULE)
class MinimalPathExtractor:
    """Minimal path extractor

    Minimal path extractor based on the fast marching method and ODE solver.

    """

    def __init__(self, speed_data: np.ndarray, source_point: PointType,
                 parameters: Optional[Parameters] = None) -> None:

        if parameters is None:  # pragma: no cover
            parameters = default_parameters()

        travel_time, phi = self._compute_travel_time(speed_data, source_point, parameters)
        gradients = np.gradient(travel_time, parameters.travel_time_spacing)

        grad_interpolants, tt_interpolant, phi_interpolant = self._compute_interpolants(gradients, travel_time, phi)

        self.speed_data = speed_data
        self.travel_time = travel_time
        self.phi = phi

        self.source_point = source_point

        self.travel_time_interpolant = tt_interpolant
        self.phi_interpolant = phi_interpolant
        self.gradient_interpolants = grad_interpolants

        self.parameters = parameters

        # output after compute ODE solution
        self.integrate_times = []
        self.path_points = []
        self.path_travel_times = []
        self.steps = 0
        self.func_eval_count = 0

    @staticmethod
    def _compute_travel_time(speed_data: np.ndarray,
                             source_point: PointType,
                             parameters: Parameters):
        # define the zero contour and set the wave source
        phi = np.ones_like(speed_data)
        phi[source_point] = -1

        try:
            travel_time = fmm.travel_time(phi, speed_data,
                                          dx=parameters.travel_time_spacing,
                                          order=parameters.travel_time_order)
        except Exception as err:  # pragma: no cover
            raise ComputeTravelTimeError from err

        return travel_time, phi

    @staticmethod
    def _compute_interpolants(gradients, travel_time, phi):
        grid_coords = [np.arange(n) for n in travel_time.shape]

        gradient_interpolants = []
        for gradient in gradients:
            interpolant = make_interpolator(grid_coords, gradient, fill_value=0.0)
            gradient_interpolants.append(interpolant)

        tt_interpolant = make_interpolator(grid_coords, travel_time, fill_value=0.0)
        phi_interpolant = make_interpolator(grid_coords, phi, fill_value=1.0)

        return gradient_interpolants, tt_interpolant, phi_interpolant

    def __call__(self, start_point: PointType) -> np.ndarray:
        gradient_interpolants = self.gradient_interpolants
        travel_time_interpolant = self.travel_time_interpolant

        def right_hand_func(time: float, point: np.ndarray) -> np.ndarray:  # noqa
            velocity = np.array([gi(point).item() for gi in gradient_interpolants])

            if np.any(np.isclose(velocity, 0.0)):
                # zero-velocity most often means masked travel time data
                return velocity

            return -velocity / np.linalg.norm(velocity)

        solver_cls = ODE_SOLVER_METHODS[self.parameters.ode_solver_method]
        logger.debug("ODE solver '%s' will be used.", solver_cls.__name__)

        with warnings.catch_warnings():
            # filter warn "extraneous arguments"
            warnings.simplefilter('ignore', category=UserWarning)

            solver = solver_cls(
                right_hand_func,
                t0=0.0,
                t_bound=self.parameters.integrate_time_bound,
                y0=start_point,
                min_step=self.parameters.integrate_min_step,
                max_step=self.parameters.integrate_max_step,
                first_step=None,
            )

        self.integrate_times = []
        self.path_travel_times = []
        self.path_points = []
        self.steps = 0

        end_point = self.source_point
        dist_tol = self.parameters.dist_tol

        y_old = start_point
        small_dist_steps_left = self.parameters.max_small_dist_steps

        while True:
            self.steps += 1
            message = solver.step()

            if solver.status == 'failed':  # pragma: no cover
                raise PathExtractionError(
                    f"ODE solver '{solver_cls.__name__}' has failed: {message}",
                    travel_time=self.travel_time, start_point=start_point, end_point=end_point)

            t = solver.t
            y = solver.y
            tt = travel_time_interpolant(y).item()

            add_point = True

            if solver.y_old is not None:
                y_old = solver.y_old
            y_dist = euclidean(y, y_old)

            if y_dist < dist_tol:
                logger.warning('step: %d, the distance between old and current extracted point (%f) is '
                               'too small (less than dist_tol=%f)', self.steps, y_dist, dist_tol)
                add_point = False
                small_dist_steps_left -= 1

            if add_point:
                small_dist_steps_left = self.parameters.max_small_dist_steps

                self.integrate_times.append(t)
                self.path_points.append(y)
                self.path_travel_times.append(tt)

            self.func_eval_count = solver.nfev

            step_size = solver.step_size
            dist_to_end = euclidean(y, end_point)

            logger.debug('step: %d, time: %.2f, point: %s, step_size: %.2f, nfev: %d, dist: %.2f, message: "%s"',
                         self.steps, t, y, step_size, solver.nfev, dist_to_end, message)

            if dist_to_end < step_size:
                logger.debug(
                    'The minimal path has been extracted (time: %.2f, steps: %d, nfev: %d, dist_to_end: %.2f)',
                    t, self.steps, solver.nfev, dist_to_end)
                break

            if solver.status == 'finished' or small_dist_steps_left == 0:
                if small_dist_steps_left == 0:
                    reason = f'the distance between old and current point stay too small ' \
                             f'for {self.parameters.max_small_dist_steps} steps'
                else:
                    reason = f'time bound {self.parameters.integrate_time_bound} is reached, solver was finished.'

                err_msg = (
                    f'The extracted path from the start point {start_point} '
                    f'did not reach the end point {end_point} in {t} time and {self.steps} steps '
                    f'with distance {dist_to_end:.2f} to the end point. Reason: {reason}'
                )

                raise EndPointNotReachedError(
                    err_msg,
                    travel_time=self.travel_time,
                    start_point=start_point,
                    end_point=end_point,
                    extracted_points=self.path_points,
                    last_distance=dist_to_end,
                    reason=reason,
                )

        return np.array(self.path_points)


def extract_path_without_way_points(init_info: InitialInfo,
                                    parameters: Parameters) -> ResultPathInfo:
    extractor = MinimalPathExtractor(init_info.speed_data, init_info.end_point, parameters)
    path = extractor(init_info.start_point)

    path_info = PathInfo(
        path=path,
        start_point=init_info.start_point,
        end_point=init_info.end_point,
        travel_time=extractor.travel_time,
        path_travel_times=np.asarray(extractor.path_travel_times),
        reversed=False,
    )

    return ResultPathInfo(path=path, pieces=[path_info])


def make_whole_path_from_pieces(path_pieces_info: List[PathInfo]) -> ResultPathInfo:
    path_pieces = [path_pieces_info[0].path]

    for path_info in path_pieces_info[1:]:
        path = path_info.path
        if path_info.reversed:
            path = np.flipud(path)
        path_pieces.append(path)

    return ResultPathInfo(
        path=np.vstack(path_pieces),
        pieces=path_pieces_info,
    )


def extract_path_with_way_points(init_info: InitialInfo,
                                 parameters: Parameters) -> ResultPathInfo:
    speed_data = init_info.speed_data
    path_pieces_info = []

    if parameters.travel_time_cache:
        compute_ttime = [True, False]
        last_extractor = None

        for (start_point, end_point), compute_tt in zip(
                init_info.point_intervals(), itertools.cycle(compute_ttime)):
            if compute_tt:
                extractor = MinimalPathExtractor(speed_data, end_point, parameters)
                last_extractor = extractor
                is_reversed = False
            else:
                extractor = last_extractor
                start_point, end_point = end_point, start_point
                is_reversed = True

            path = extractor(start_point)

            path_pieces_info.append(PathInfo(
                path=path,
                start_point=start_point,
                end_point=end_point,
                travel_time=extractor.travel_time,
                path_travel_times=np.asarray(extractor.path_travel_times),
                reversed=is_reversed
            ))
    else:
        for start_point, end_point in init_info.point_intervals():
            extractor = MinimalPathExtractor(speed_data, end_point, parameters)
            path = extractor(start_point)

            path_pieces_info.append(PathInfo(
                path=path,
                start_point=start_point,
                end_point=end_point,
                travel_time=extractor.travel_time,
                path_travel_times=np.asarray(extractor.path_travel_times),
                reversed=False
            ))

    return make_whole_path_from_pieces(path_pieces_info)


def extract_path(init_info: InitialInfo,
                 parameters: Optional[Parameters] = None) -> ResultPathInfo:

    if parameters is None:
        parameters = default_parameters()

    if init_info.way_points:
        return extract_path_with_way_points(init_info, parameters)
    else:
        return extract_path_without_way_points(init_info, parameters)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['skmpe']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.18.1,<2.0.0',
 'pydantic>=1.4,<2.0',
 'scikit-fmm>=2019.1.30,<2020.0.0',
 'scipy>=1.4.1,<2.0.0']

extras_require = \
{'docs': ['sphinx>=2.3.1,<3.0.0',
          'matplotlib>=3.1.3,<4.0.0',
          'numpydoc>=0.9.2,<0.10.0',
          'm2r>=0.2.1,<0.3.0',
          'toml>=0.10.0,<0.11.0'],
 'tests': ['pytest>=5.3.5,<6.0.0',
           'pytest-cov>=2.8.1,<3.0.0',
           'coveralls>=1.10.0,<2.0.0',
           'scikit-image>=0.16.2,<0.17.0']}

setup_kwargs = {
    'name': 'scikit-mpe',
    'version': '0.1.1',
    'description': 'Minimal path extraction using the fast marching method',
    'long_description': "# scikit-mpe\n\n[![PyPI version](https://img.shields.io/pypi/v/scikit-mpe.svg)](https://pypi.python.org/pypi/scikit-mpe)\n[![Build status](https://travis-ci.org/espdev/scikit-mpe.svg?branch=master)](https://travis-ci.org/espdev/scikit-mpe)\n[![Documentation Status](https://readthedocs.org/projects/scikit-mpe/badge/?version=latest)](https://scikit-mpe.readthedocs.io/en/latest/?badge=latest)\n[![Coverage Status](https://coveralls.io/repos/github/espdev/scikit-mpe/badge.svg?branch=master)](https://coveralls.io/github/espdev/scikit-mpe?branch=master)\n![Supported Python versions](https://img.shields.io/pypi/pyversions/scikit-mpe.svg)\n[![License](https://img.shields.io/pypi/l/scikit-mpe.svg)](LICENSE)\n\n**scikit-mpe** is a package for extracting a minimal path in N-dimensional Euclidean space (on regular Cartesian grids) \nusing [the fast marching method](https://math.berkeley.edu/~sethian/2006/Explanations/fast_marching_explain.html).\n\n\n## Quickstart\n\n### Installing\n\n```\npip install -U scikit-mpe\n```\n\n### Usage\n\nHere is a simple example that demonstrates how to extract the path passing through the retina vessels.\n\n```python\nfrom matplotlib import pyplot as plt\n\nfrom skimage.data import retina\nfrom skimage.color import rgb2gray\nfrom skimage.transform import rescale\nfrom skimage.filters import sato\n\nfrom skmpe import mpe\n\nimage = rescale(rgb2gray(retina()), 0.5)\nspeed_image = sato(image)\n\nstart_point = (76, 388)\nend_point = (611, 442)\nway_points = [(330, 98), (554, 203)]\n\npath_info = mpe(speed_image, start_point, end_point, way_points)\n\npx, py = path_info.path[:, 1], path_info.path[:, 0]\n\nplt.imshow(image, cmap='gray')\nplt.plot(px, py, '-r')\n\nplt.plot(*start_point[::-1], 'oy')\nplt.plot(*end_point[::-1], 'og')\nfor p in way_points:\n    plt.plot(*p[::-1], 'ob')\n\nplt.show()\n```\n\n![retina_vessel_path](https://user-images.githubusercontent.com/1299189/73838143-0d74c380-4824-11ea-946a-667c8236ed75.png)\n\n## Documentation\n\nThe full documentation can be found at https://scikit-mpe.readthedocs.io/en/latest\n\n(The documentation is being written)\n\n## References\n\n- [Fast Marching Methods: A boundary value formulation](https://math.berkeley.edu/~sethian/2006/Explanations/fast_marching_explain.html)\n- [Level Set Methods and Fast Marching Methods](https://math.berkeley.edu/~sethian/2006/History/Menu_Expanded_History.html)\n- [scikit-fmm](https://github.com/scikit-fmm/scikit-fmm) - Python implementation of the fast marching method\n- [ITKMinimalPathExtraction](https://github.com/InsightSoftwareConsortium/ITKMinimalPathExtraction) - ITK based C++ implementation of MPE\n\n## License\n\n[MIT](https://choosealicense.com/licenses/mit/)\n",
    'author': 'Eugene Prilepin',
    'author_email': 'esp.home@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/espdev/scikit-mpe',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
